@@ -10,6 +10,7 @@
 #include "text.h"
 #include "expr.h"
 #include "keysym.h"
+#include "xkbcomp/ast.h"
 
 typedef bool (*IdentLookupFunc)(struct xkb_context *ctx, const void *priv,
                                 xkb_atom_t field, uint32_t *val_rtrn);
@@ -43,8 +44,8 @@ ExprResolveLhs(struct xkb_context *ctx, const ExprDef *expr,
         break;
     }
     log_wsgo(ctx,
-        XKB_ERROR_INVALID_SYNTAX,
-        "Unexpected operator %d in ResolveLhs\n", expr->common.type);
+             XKB_ERROR_INVALID_XKB_SYNTAX,
+             "Unexpected operator %d in ResolveLhs\n", expr->common.type);
     return false;
 }
 
@@ -117,6 +118,7 @@ ExprResolveBoolean(struct xkb_context *ctx, const ExprDef *expr,
     case STMT_EXPR_INTEGER_LITERAL:
     case STMT_EXPR_FLOAT_LITERAL:
     case STMT_EXPR_KEYNAME_LITERAL:
+    case STMT_EXPR_KEYSYM_LITERAL:
         log_err(ctx, XKB_ERROR_WRONG_FIELD_TYPE,
                 "Found %s where boolean was expected\n",
                 stmt_type_to_string(expr->common.type));
@@ -200,6 +202,7 @@ ExprResolveIntegerLookup(struct xkb_context *ctx, const ExprDef *expr,
     case STMT_EXPR_FLOAT_LITERAL:
     case STMT_EXPR_BOOLEAN_LITERAL:
     case STMT_EXPR_KEYNAME_LITERAL:
+    case STMT_EXPR_KEYSYM_LITERAL:
         log_err(ctx, XKB_ERROR_WRONG_FIELD_TYPE,
                 "Found %s where an int was expected\n",
                 stmt_type_to_string(expr->common.type));
@@ -367,6 +370,7 @@ ExprResolveString(struct xkb_context *ctx, const ExprDef *expr,
     case STMT_EXPR_FLOAT_LITERAL:
     case STMT_EXPR_BOOLEAN_LITERAL:
     case STMT_EXPR_KEYNAME_LITERAL:
+    case STMT_EXPR_KEYSYM_LITERAL:
         log_err(ctx, XKB_ERROR_WRONG_FIELD_TYPE,
                 "Found %s, expected a string\n",
                 stmt_type_to_string(expr->common.type));
@@ -398,7 +402,7 @@ ExprResolveString(struct xkb_context *ctx, const ExprDef *expr,
     case STMT_EXPR_ACTION_DECL:
     case STMT_EXPR_ACTION_LIST:
     case STMT_EXPR_KEYSYM_LIST:
-        log_err(ctx, XKB_ERROR_INVALID_SYNTAX,
+        log_err(ctx, XKB_ERROR_INVALID_XKB_SYNTAX,
                 "%s of strings not permitted\n",
                 stmt_type_to_string(expr->common.type));
         return false;
@@ -466,6 +470,7 @@ ExprResolveMaskLookup(struct xkb_context *ctx, const ExprDef *expr,
     case STMT_EXPR_FLOAT_LITERAL:
     case STMT_EXPR_BOOLEAN_LITERAL:
     case STMT_EXPR_KEYNAME_LITERAL:
+    case STMT_EXPR_KEYSYM_LITERAL:
         log_err(ctx, XKB_ERROR_WRONG_FIELD_TYPE,
                 "Found %s where a mask was expected\n",
                 stmt_type_to_string(expr->common.type));
@@ -581,53 +586,6 @@ ExprResolveModMask(struct xkb_context *ctx, const ExprDef *expr,
     return ExprResolveMaskLookup(ctx, expr, mask_rtrn, LookupModMask, &priv);
 }
 
-bool
-ExprResolveKeySym(struct xkb_context *ctx, const ExprDef *expr,
-                  xkb_keysym_t *sym_rtrn)
-{
-    int64_t val = 0;
-
-    if (expr->common.type == STMT_EXPR_IDENT) {
-        const char *str = xkb_atom_text(ctx, expr->ident.ident);
-        *sym_rtrn = xkb_keysym_from_name(str, 0);
-        if (*sym_rtrn != XKB_KEY_NoSymbol) {
-            check_deprecated_keysyms(log_warn, ctx, ctx,
-                                     *sym_rtrn, str, str, "%s", "\n");
-            return true;
-        }
-    }
-
-    if (!ExprResolveInteger(ctx, expr, &val))
-        return false;
-
-    if (val < XKB_KEYSYM_MIN) {
-        log_warn(ctx, XKB_WARNING_UNRECOGNIZED_KEYSYM,
-                 "unrecognized keysym \"-0x%"PRIx64"\" (%"PRId64")\n",
-                 (uint64_t) -val, val);
-        return false;
-    }
-
-    /* Special case for digits 0..9 */
-    if (val < 10) {
-        *sym_rtrn = XKB_KEY_0 + (xkb_keysym_t) val;
-        return true;
-    }
-
-    if (val <= XKB_KEYSYM_MAX) {
-        check_deprecated_keysyms(log_warn, ctx, ctx, val, NULL, val, "0x%"PRIx64, "\n");
-        log_warn(ctx, XKB_WARNING_NUMERIC_KEYSYM,
-                 "numeric keysym \"0x%"PRIx64"\" (%"PRId64")",
-                 (uint64_t) val, val);
-        *sym_rtrn = (xkb_keysym_t) val;
-        return true;
-    }
-
-    log_warn(ctx, XKB_WARNING_UNRECOGNIZED_KEYSYM,
-             "unrecognized keysym \"0x%"PRIx64"\" (%"PRId64")\n",
-             (uint64_t) val, val);
-    return false;
-}
-
 bool
 ExprResolveMod(struct xkb_context *ctx, const ExprDef *def,
                enum mod_type mod_type, const struct xkb_mod_set *mods,
