@@ -10,12 +10,16 @@
 
 #include "config.h"
 
+#include <assert.h>
 #include <limits.h>
 
+#include "darray.h"
+#include "xkbcommon/xkbcommon.h"
 #include "xkbcomp-priv.h"
 #include "rules.h"
 #include "include.h"
 #include "scanner-utils.h"
+#include "utils-numbers.h"
 #include "utils-paths.h"
 
 #define MAX_INCLUDE_DEPTH 5
@@ -34,7 +38,10 @@ enum rules_token {
     TOK_GROUP_NAME,
     TOK_BANG,
     TOK_EQUALS,
-    TOK_STAR,
+    TOK_WILD_CARD_STAR,
+    TOK_WILD_CARD_NONE,
+    TOK_WILD_CARD_SOME,
+    TOK_WILD_CARD_ANY,
     TOK_INCLUDE,
     TOK_ERROR
 };
@@ -68,7 +75,7 @@ skip_more_whitespace_and_comments:
         /* Optional \r. */
         scanner_chr(s, '\r');
         if (!scanner_eol(s)) {
-            scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+            scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                         "illegal new line escape; must appear at end of line");
             return TOK_ERROR;
         }
@@ -85,7 +92,12 @@ skip_more_whitespace_and_comments:
     /* Operators and punctuation. */
     if (scanner_chr(s, '!')) return TOK_BANG;
     if (scanner_chr(s, '=')) return TOK_EQUALS;
-    if (scanner_chr(s, '*')) return TOK_STAR;
+
+    /* Wild cards */
+    if (scanner_chr(s, '*')) return TOK_WILD_CARD_STAR;
+    if (scanner_lit(s, "<none>")) return TOK_WILD_CARD_NONE;
+    if (scanner_lit(s, "<some>")) return TOK_WILD_CARD_SOME;
+    if (scanner_lit(s, "<any>")) return TOK_WILD_CARD_ANY;
 
     /* Group name. */
     if (scanner_chr(s, '$')) {
@@ -96,7 +108,7 @@ skip_more_whitespace_and_comments:
             val->string.len++;
         }
         if (val->string.len == 0) {
-            scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+            scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                         "unexpected character after \'$\'; expected name");
             return TOK_ERROR;
         }
@@ -108,6 +120,10 @@ skip_more_whitespace_and_comments:
         return TOK_INCLUDE;
 
     /* Identifier. */
+    /* Ensure that we can parse KcCGST values with merge modes */
+    assert(is_ident(MERGE_OVERRIDE_PREFIX));
+    assert(is_ident(MERGE_AUGMENT_PREFIX));
+    assert(is_ident(MERGE_REPLACE_PREFIX));
     if (is_ident(scanner_peek(s))) {
         val->string.start = s->s + s->pos;
         val->string.len = 0;
@@ -118,7 +134,7 @@ skip_more_whitespace_and_comments:
         return TOK_IDENTIFIER;
     }
 
-    scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+    scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                 "unrecognized token");
     return TOK_ERROR;
 }
@@ -208,15 +224,24 @@ struct mapping {
 };
 
 enum mlvo_match_type {
+    /** Match the given plain value */
     MLVO_MATCH_NORMAL = 0,
-    MLVO_MATCH_WILDCARD,
+    /** Match depending on the value of `wildcard_match_type` */
+    MLVO_MATCH_WILDCARD_LEGACY,
+    /** Match empty value */
+    MLVO_MATCH_WILDCARD_NONE,
+    /** Match non-empty value */
+    MLVO_MATCH_WILDCARD_SOME,
+    /** Match any value, optionally empty */
+    MLVO_MATCH_WILDCARD_ANY,
+    /** Match any entry in a group */
     MLVO_MATCH_GROUP,
 };
 
 enum wildcard_match_type {
-    /* Match only non-empty strings */
+    /** ‘*’ matches only non-empty strings */
     WILDCARD_MATCH_NONEMPTY = 0,
-    /* Match all strings */
+    /** ‘*’ matches all strings */
     WILDCARD_MATCH_ALL,
 };
 
@@ -229,6 +254,23 @@ struct rule {
     bool skip;
 };
 
+#define SLICE_KCCGST_BIT_FIELD_SIZE 4
+static_assert(!((1u << (SLICE_KCCGST_BIT_FIELD_SIZE - 1)) &
+                (_KCCGST_NUM_ENTRIES - 1)),
+              "Cannot encode KcCGST enum safely (need space for the sign)");
+struct kccgst_buffer_slice {
+    uint32_t length:(32 - SLICE_KCCGST_BIT_FIELD_SIZE);
+    enum rules_kccgst kccgst:SLICE_KCCGST_BIT_FIELD_SIZE;
+    xkb_layout_index_t layout;
+};
+
+/* Buffer for pending KcCGST values */
+struct kccgst_buffer {
+    darray_char buffer;
+    /* Slice corresponding to each value in the buffer */
+    darray(struct kccgst_buffer_slice) slices;
+};
+
 /*
  * This is the main object used to match a given RMLVO against a rules
  * file and aggregate the results in a KcCGST. It goes through a simple
@@ -245,6 +287,12 @@ struct matcher {
     struct mapping mapping;
     /* Current rule. */
     struct rule rule;
+    /*
+     * Buffers for pending KcCGST values. Required in case of using layout
+     * index ranges, to ensure that the values are merged in the expected order.
+     * See the note: “Layout index ranges and merging KcCGST values”.
+     */
+    struct kccgst_buffer pending_kccgst;
     /* Output. */
     darray_char kccgst[_KCCGST_NUM_ENTRIES];
 };
@@ -323,14 +371,16 @@ matcher_new(struct xkb_context *ctx,
 static void
 matcher_free(struct matcher *m)
 {
-    struct group *group;
     if (!m)
         return;
     darray_free(m->rmlvo.layouts);
     darray_free(m->rmlvo.variants);
     darray_free(m->rmlvo.options);
+    struct group *group;
     darray_foreach(group, m->groups)
         darray_free(group->elements);
+    darray_free(m->pending_kccgst.buffer);
+    darray_free(m->pending_kccgst.slices);
     for (int i = 0; i < _KCCGST_NUM_ENTRIES; i++)
         darray_free(m->kccgst[i]);
     darray_free(m->groups);
@@ -496,16 +546,19 @@ extract_layout_index(const char *s, size_t max_len, xkb_layout_index_t *out)
     }
     /* Numeric index */
 
-#define parse_layout_int_index(s, endptr, index, out)                  \
-    char *(endptr);                                                    \
-    long (index) = strtol(&(s)[1], &(endptr), 10);                     \
-    if ((index) < 0 || (endptr)[0] != ']' || (index) > XKB_MAX_GROUPS) \
-        return -1;                                                     \
-    /* To zero-based index. */                                         \
-    *(out) = (xkb_layout_index_t)(index) - 1;                          \
-    return (int)((endptr) - (s) + 1) /* == length "[index]" */
+#define parse_layout_int_index(s, len, out) do {                                \
+    /* We expect a NULL-terminated string of at least length 3 */               \
+    assert((len) >= 3);                                                         \
+    uint32_t val = 0;                                                           \
+    const int count = parse_hex_to_uint32_t(&(s)[1], (len) - 2, &val);          \
+    if (count <= 0 || (s)[1 + count] != ']' || val == 0 || val > XKB_MAX_GROUPS)\
+        return -1;                                                              \
+    /* To zero-based index. */                                                  \
+    *(out) = val - 1;                                                           \
+    return count + 2; /* == length "[index]" */                                 \
+} while (0)
 
-    parse_layout_int_index(s, endptr, index, out);
+    parse_layout_int_index(s, max_len, out);
 }
 
 /* Special layout indexes */
@@ -549,7 +602,7 @@ extract_mapping_layout_index(const char *s, size_t max_len,
     else if_index(&s[1], LAYOUT_INDEX_ANY, out)
     else {
         /* Numeric index */
-        parse_layout_int_index(s, endptr, index, out);
+        parse_layout_int_index(s, max_len, out);
     }
 #undef if_index
 #undef LAYOUT_INDEX_SINGLE
@@ -574,8 +627,8 @@ matcher_mapping_set_mlvo(struct matcher *m, struct scanner *s,
 
     /* Not found. */
     if (mlvo >= _MLVO_NUM_ENTRIES) {
-        scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
-                    "invalid mapping: %.*s is not a valid value here; "
+        scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
+                    "invalid mapping: \"%.*s\" is not a valid value here; "
                     "ignoring rule set",
                     ident.len, ident.start);
         m->mapping.active = false;
@@ -583,8 +636,8 @@ matcher_mapping_set_mlvo(struct matcher *m, struct scanner *s,
     }
 
     if (is_mlvo_mask_defined(m, mlvo)) {
-        scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
-                    "invalid mapping: %.*s appears twice on the same line; "
+        scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
+                    "invalid mapping: \"%.*s\" appears twice on the same line; "
                     "ignoring rule set",
                     mlvo_sval.len, mlvo_sval.start);
         m->mapping.active = false;
@@ -598,7 +651,7 @@ matcher_mapping_set_mlvo(struct matcher *m, struct scanner *s,
                                                     ident.len - mlvo_sval.len,
                                                     &idx);
         if ((int) (ident.len - mlvo_sval.len) != consumed) {
-            scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+            scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                         "invalid mapping: \"%.*s\" may only be followed by a "
                         "valid group index; ignoring rule set",
                         mlvo_sval.len, mlvo_sval.start);
@@ -613,7 +666,7 @@ matcher_mapping_set_mlvo(struct matcher *m, struct scanner *s,
             m->mapping.variant_idx = idx;
         }
         else {
-            scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+            scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                         "invalid mapping: \"%.*s\" cannot be followed by a group "
                         "index; ignoring rule set",
                         mlvo_sval.len, mlvo_sval.start);
@@ -627,7 +680,7 @@ matcher_mapping_set_mlvo(struct matcher *m, struct scanner *s,
     if (((mlvo == MLVO_LAYOUT && is_mlvo_mask_defined(m, MLVO_VARIANT)) ||
          (mlvo == MLVO_VARIANT && is_mlvo_mask_defined(m, MLVO_LAYOUT))) &&
         m->mapping.layout_idx != m->mapping.variant_idx) {
-        scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+        scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                     "invalid mapping: \"layout\" index must be the same as the "
                     "\"variant\" index");
         m->mapping.active = false;
@@ -693,8 +746,8 @@ matcher_mapping_set_kccgst(struct matcher *m, struct scanner *s, struct sval ide
 
     /* Not found. */
     if (kccgst >= _KCCGST_NUM_ENTRIES) {
-        scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
-                    "invalid mapping: %.*s is not a valid value here; "
+        scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
+                    "invalid mapping: \"%.*s\" is not a valid value here; "
                     "ignoring rule set",
                     ident.len, ident.start);
         m->mapping.active = false;
@@ -702,8 +755,8 @@ matcher_mapping_set_kccgst(struct matcher *m, struct scanner *s, struct sval ide
     }
 
     if (m->mapping.defined_kccgst_mask & (1u << kccgst)) {
-        scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
-                    "invalid mapping: %.*s appears twice on the same line; "
+        scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
+                    "invalid mapping: \"%.*s\" appears twice on the same line; "
                     "ignoring rule set",
                     kccgst_sval.len, kccgst_sval.start);
         m->mapping.active = false;
@@ -719,14 +772,14 @@ static bool
 matcher_mapping_verify(struct matcher *m, struct scanner *s)
 {
     if (m->mapping.num_mlvo == 0) {
-        scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+        scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                     "invalid mapping: must have at least one value on the left "
                     "hand side; ignoring rule set");
         goto skip;
     }
 
     if (m->mapping.num_kccgst == 0) {
-        scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+        scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                     "invalid mapping: must have at least one value on the right "
                     "hand side; ignoring rule set");
         goto skip;
@@ -801,7 +854,7 @@ matcher_rule_set_mlvo_common(struct matcher *m, struct scanner *s,
                              enum mlvo_match_type match_type)
 {
     if (m->rule.num_mlvo_values + 1 > m->mapping.num_mlvo) {
-        scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+        scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                     "invalid rule: has more values than the mapping line; "
                     "ignoring rule");
         m->rule.skip = true;
@@ -813,10 +866,11 @@ matcher_rule_set_mlvo_common(struct matcher *m, struct scanner *s,
 }
 
 static void
-matcher_rule_set_mlvo_wildcard(struct matcher *m, struct scanner *s)
+matcher_rule_set_mlvo_wildcard(struct matcher *m, struct scanner *s,
+                               enum mlvo_match_type match_type)
 {
     struct sval dummy = { NULL, 0 };
-    matcher_rule_set_mlvo_common(m, s, dummy, MLVO_MATCH_WILDCARD);
+    matcher_rule_set_mlvo_common(m, s, dummy, match_type);
 }
 
 static void
@@ -838,7 +892,7 @@ matcher_rule_set_kccgst(struct matcher *m, struct scanner *s,
                         struct sval ident)
 {
     if (m->rule.num_kccgst_values + 1 > m->mapping.num_kccgst) {
-        scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+        scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                     "invalid rule: has more values than the mapping line; "
                     "ignoring rule");
         m->rule.skip = true;
@@ -883,12 +937,23 @@ match_value(struct matcher *m, struct sval val, struct sval to,
             enum mlvo_match_type match_type,
             enum wildcard_match_type wildcard_type)
 {
-    if (match_type == MLVO_MATCH_WILDCARD)
-        /* Wild card match empty values only if explicitly required */
-        return wildcard_type == WILDCARD_MATCH_ALL || !!to.len;
-    if (match_type == MLVO_MATCH_GROUP)
-        return match_group(m, val, to);
-    return svaleq(val, to);
+    switch (match_type) {
+        case MLVO_MATCH_WILDCARD_LEGACY:
+            /* Match empty values only if explicitly required */
+            return wildcard_type == WILDCARD_MATCH_ALL || !!to.len;
+        case MLVO_MATCH_WILDCARD_NONE:
+            return !to.len;
+        case MLVO_MATCH_WILDCARD_SOME:
+            return !!to.len;
+        case MLVO_MATCH_WILDCARD_ANY:
+            /* Contrary to the legacy ‘*’, this wild card *always* matches */
+            return true;
+        case MLVO_MATCH_GROUP:
+            return match_group(m, val, to);
+        default:
+            assert(match_type == MLVO_MATCH_NORMAL);
+            return svaleq(val, to);
+    }
 }
 
 static bool
@@ -958,7 +1023,7 @@ expand_rmlvo_in_kccgst_value(struct matcher *m, struct scanner *s,
     bool expanded_index = false;
     if (*i < value.len && str[*i] == '[') {
         if (mlv != MLVO_LAYOUT && mlv != MLVO_VARIANT) {
-            scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+            scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                         "invalid index in %%-expansion; "
                         "may only index layout or variant");
             goto error;
@@ -1028,7 +1093,7 @@ expand_rmlvo_in_kccgst_value(struct matcher *m, struct scanner *s,
     return true;
 
 error:
-    scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+    scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                 "invalid %%-expansion in value; not used");
     return false;
 }
@@ -1080,19 +1145,42 @@ expand_qualifier_in_kccgst_value(
     }
 }
 
+static inline void
+#ifdef _MSC_VER
+concat_kccgst(darray_char *into, size_t size, _In_reads_(size) const char* from)
+#else
+concat_kccgst(darray_char *into, size_t size, const char from[static size])
+#endif
+{
+    /*
+     * Appending  bar to  foo ->  foo (not an error if this happens)
+     * Appending +bar to  foo ->  foo+bar
+     * Appending  bar to +foo ->  bar+foo
+     * Appending +bar to +foo -> +foo+bar
+     */
+    const bool from_plus = is_merge_mode_prefix(from[0]);
+    if (from_plus || darray_empty(*into)) {
+        darray_appends_nullterminate(*into, from, size);
+    } else {
+        const char ch =
+            (char) (darray_empty(*into) ? '\0' : darray_item(*into, 0));
+        const bool into_plus = is_merge_mode_prefix(ch);
+        if (into_plus)
+            darray_prepends_nullterminate(*into, from, size);
+    }
+}
+
 /*
  * This function performs %-expansion and :all-expansion on @value
  * (see overview above), and appends the result to @to.
  */
 static bool
 append_expanded_kccgst_value(struct matcher *m, struct scanner *s,
-                             darray_char *to, struct sval value,
+                             bool merge, darray_char *to, struct sval value,
                              xkb_layout_index_t layout_idx)
 {
     const char *str = value.start;
     darray_char expanded = darray_new();
-    char ch;
-    bool expanded_plus, to_plus;
     unsigned int last_item_idx = 0;
     bool has_separator = false;
 
@@ -1118,6 +1206,7 @@ append_expanded_kccgst_value(struct matcher *m, struct scanner *s,
             /* New item */
             case MERGE_OVERRIDE_PREFIX:
             case MERGE_AUGMENT_PREFIX:
+            case MERGE_REPLACE_PREFIX:
                 darray_appends_nullterminate(expanded, &str[i++], 1);
                 last_item_idx = darray_size(expanded) - 1;
                 has_separator = true;
@@ -1128,27 +1217,13 @@ append_expanded_kccgst_value(struct matcher *m, struct scanner *s,
         }
     }
 
-    /*
-     * Appending  bar to  foo ->  foo (not an error if this happens)
-     * Appending +bar to  foo ->  foo+bar
-     * Appending  bar to +foo ->  bar+foo
-     * Appending +bar to +foo -> +foo+bar
-     */
-
-    ch = (char) (darray_empty(expanded) ? '\0' : darray_item(expanded, 0));
-    expanded_plus = is_merge_mode_prefix(ch);
-    ch = (char) (darray_empty(*to) ? '\0' : darray_item(*to, 0));
-    to_plus = is_merge_mode_prefix(ch);
-
-    if (expanded_plus || darray_empty(*to))
-        darray_appends_nullterminate(*to,
-                                     darray_items(expanded),
-                                     darray_size(expanded));
-    else if (to_plus)
-        darray_prepends_nullterminate(*to,
-                                      darray_items(expanded),
-                                      darray_size(expanded));
-
+    /* See note: “Layout index ranges and merging KcCGST values” */
+    if (merge) {
+        if (!darray_empty(expanded))
+            concat_kccgst(to, darray_size(expanded), darray_items(expanded));
+    } else {
+        darray_concat(*to, expanded);
+    }
     darray_free(expanded);
     return true;
 
@@ -1157,12 +1232,48 @@ error:
     return false;
 }
 
+static bool
+matcher_append_pending_kccgst(struct matcher *m)
+{
+    if (!m->mapping.has_layout_idx_range)
+        return true;
+    /*
+     * Handle pending KcCGST values
+     * See note: “Layout index ranges and merging KcCGST values”
+     */
+    for (unsigned int i = 0; i < m->mapping.num_kccgst; i++) {
+        const enum rules_kccgst kccgst = m->mapping.kccgst_at_pos[i];
+        /* For each relevant layout, append the relevant KcCGST values to
+         * the output. */
+        for (xkb_layout_index_t layout = m->mapping.layout_idx_min;
+             layout < m->mapping.layout_idx_max;
+             layout++) {
+            /* There may be multiple values to add if the rule set involved
+             * options. Process them sequentially. */
+            register const struct kccgst_buffer* const buf = &m->pending_kccgst;
+            size_t offset = 0;
+            for (unsigned k = 0; k < darray_size(buf->slices); k++) {
+                register const struct kccgst_buffer_slice * const slice =
+                    &darray_item(buf->slices, k);
+                if (slice->kccgst == kccgst && slice->layout == layout &&
+                    slice->length)
+                    concat_kccgst(&m->kccgst[kccgst], slice->length,
+                                  darray_items(buf->buffer) + offset);
+                offset += slice->length;
+            }
+        }
+    }
+    /* Ensure we won’t come here before the next relevant rule set */
+    m->mapping.has_layout_idx_range = false;
+    return true;
+}
+
 static void
 matcher_rule_verify(struct matcher *m, struct scanner *s)
 {
     if (m->rule.num_mlvo_values != m->mapping.num_mlvo ||
         m->rule.num_kccgst_values != m->mapping.num_kccgst) {
-        scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+        scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                     "invalid rule: must have same number of values "
                     "as mapping line; ignoring rule");
         m->rule.skip = true;
@@ -1183,7 +1294,7 @@ matcher_rule_apply_if_matches(struct matcher *m, struct scanner *s)
         struct matched_sval *to;
         bool matched = false;
 
-        /* NOTE: Wild card matches empty values only for model and options, as
+        /* NOTE: Wild card * matches empty values only for model and options, as
          * implemented in libxkbfile and xserver. The reason for such different
          * treatment is not documented. */
         if (mlvo == MLVO_MODEL) {
@@ -1218,7 +1329,7 @@ matcher_rule_apply_if_matches(struct matcher *m, struct scanner *s)
                 (to) = &darray_item((m)->rmlvo._component,                   \
                                   (m)->mapping.layout_idx_min);              \
                 (matched) = match_value_and_mark(m, value, to, match_type,   \
-                                               WILDCARD_MATCH_NONEMPTY);     \
+                                                 WILDCARD_MATCH_NONEMPTY);   \
             }
         else if (mlvo == MLVO_LAYOUT) {
             process_component(layouts, m, value, idx, candidate_layouts, to,
@@ -1250,10 +1361,58 @@ matcher_rule_apply_if_matches(struct matcher *m, struct scanner *s)
         {
             if (candidate_layouts & (UINT32_C(1) << idx)) {
                 for (unsigned i = 0; i < m->mapping.num_kccgst; i++) {
-                    enum rules_kccgst kccgst = m->mapping.kccgst_at_pos[i];
-                    struct sval value = m->rule.kccgst_value_at_pos[i];
-                    append_expanded_kccgst_value(m, s, &m->kccgst[kccgst],
+                    const enum rules_kccgst kccgst = m->mapping.kccgst_at_pos[i];
+                    const struct sval value = m->rule.kccgst_value_at_pos[i];
+                    /*
+                     * [NOTE] Layout index ranges and merging KcCGST values
+                     *
+                     * Layout indexes match following first the order of the
+                     * rules in the file, then their natural order. So do not
+                     * merge with the output for now but buffer the resulting
+                     * KcCGST value and wait reaching the end of the rule set.
+                     *
+                     * Because the rule set may also involve options, it may
+                     * match multiple times for the *same* layout index. So
+                     * buffer the result of *each* match.
+                     *
+                     * When the end of the rule set is reached, merge buffered
+                     * KcCGST sequentially, following first the layouts order,
+                     * then the order of the rules in the file.
+                     *
+                     * Example:
+                     *
+                     * ! model = symbols
+                     *   *     = pc
+                     * ! layout[any] option = symbols
+                     *   C           1      = +c1:%i
+                     *   C           2      = +c2:%i
+                     *   B           3      = skip
+                     *   B           4      = +b:%i
+                     *
+                     * The result of {layout: "A,B,C", options: "4,3,2,1"} is:
+                     * symbols = pc+b:2+c1:3+c2:3.
+                     *
+                     * - `skip` was dropped because it has no explicit merge
+                     *   mode;
+                     * - although every rule was matched in order, the resulting
+                     *   order of the symbols follows the order of the layouts,
+                     *   so `+b` appears before `+c1` and `+c2`.
+                     * - the relative order of the options for layout C follows
+                     *   the order within the rule set, not the order of RMLVO.
+                     */
+                    register struct kccgst_buffer * const buf =
+                        &m->pending_kccgst;
+                    const size_t prev_buffer_length = darray_size(buf->buffer);
+                    append_expanded_kccgst_value(m, s, false, &buf->buffer,
                                                  value, idx);
+                    const uint32_t length = (uint32_t) (darray_size(buf->buffer)
+                                          - prev_buffer_length);
+                    const struct kccgst_buffer_slice slice = {
+                        .length = length,
+                        .kccgst = kccgst,
+                        .layout = idx
+                    };
+                    darray_append(buf->slices, slice);
                 }
             }
         }
@@ -1262,7 +1421,7 @@ matcher_rule_apply_if_matches(struct matcher *m, struct scanner *s)
         for (unsigned i = 0; i < m->mapping.num_kccgst; i++) {
             enum rules_kccgst kccgst = m->mapping.kccgst_at_pos[i];
             struct sval value = m->rule.kccgst_value_at_pos[i];
-            append_expanded_kccgst_value(m, s, &m->kccgst[kccgst], value,
+            append_expanded_kccgst_value(m, s, true, &m->kccgst[kccgst], value,
                                          m->mapping.layout_idx_min);
         }
     }
@@ -1376,8 +1535,15 @@ mapping_kccgst:
             matcher_mapping_set_kccgst(m, s, m->val.string);
         goto mapping_kccgst;
     case TOK_END_OF_LINE:
-        if (m->mapping.active && matcher_mapping_verify(m, s))
+        if (m->mapping.active && matcher_mapping_verify(m, s)) {
             matcher_mapping_set_layout_bounds(m);
+            if (m->mapping.has_layout_idx_range) {
+                /* Lazily reset buffers for layout index ranges.
+                 * We’ll reuse the allocations. */
+                darray_size(m->pending_kccgst.buffer) = 0;
+                darray_size(m->pending_kccgst.slices) = 0;
+            }
+        }
         goto rule_mlvo_first;
     default:
         goto unexpected;
@@ -1386,10 +1552,12 @@ mapping_kccgst:
 rule_mlvo_first:
     switch (tok = gettok(m, s)) {
     case TOK_BANG:
+        matcher_append_pending_kccgst(m);
         goto bang;
     case TOK_END_OF_LINE:
         goto rule_mlvo_first;
     case TOK_END_OF_FILE:
+        matcher_append_pending_kccgst(m);
         goto finish;
     default:
         matcher_rule_start_new(m);
@@ -1401,12 +1569,28 @@ rule_mlvo:
 rule_mlvo_no_tok:
     switch (tok) {
     case TOK_IDENTIFIER:
-        if (!m->rule.skip)
-            matcher_rule_set_mlvo(m, s, m->val.string);
+        if (!m->rule.skip) {
+            if (m->val.string.len == 1 && m->val.string.start[0] == '+')
+                matcher_rule_set_mlvo_wildcard(m, s, MLVO_MATCH_WILDCARD_SOME);
+            else
+                matcher_rule_set_mlvo(m, s, m->val.string);
+        }
         goto rule_mlvo;
-    case TOK_STAR:
+    case TOK_WILD_CARD_STAR:
         if (!m->rule.skip)
-            matcher_rule_set_mlvo_wildcard(m, s);
+            matcher_rule_set_mlvo_wildcard(m, s, MLVO_MATCH_WILDCARD_LEGACY);
+        goto rule_mlvo;
+    case TOK_WILD_CARD_NONE:
+        if (!m->rule.skip)
+            matcher_rule_set_mlvo_wildcard(m, s, MLVO_MATCH_WILDCARD_NONE);
+        goto rule_mlvo;
+    case TOK_WILD_CARD_SOME:
+        if (!m->rule.skip)
+            matcher_rule_set_mlvo_wildcard(m, s, MLVO_MATCH_WILDCARD_SOME);
+        goto rule_mlvo;
+    case TOK_WILD_CARD_ANY:
+        if (!m->rule.skip)
+            matcher_rule_set_mlvo_wildcard(m, s, MLVO_MATCH_WILDCARD_ANY);
         goto rule_mlvo;
     case TOK_GROUP_NAME:
         if (!m->rule.skip)
@@ -1446,7 +1630,7 @@ finish:
     return true;
 
 state_error:
-    scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+    scanner_err(s, XKB_ERROR_INVALID_RULES_SYNTAX,
                 "unexpected token");
 error:
     return false;
@@ -1477,12 +1661,12 @@ read_rules_file(struct xkb_context *ctx,
        The first character relevant to the grammar must be ASCII:
        whitespace, !, / (for comment) */
     if (!scanner_check_supported_char_encoding(&scanner)) {
-        scanner_err(&scanner, XKB_LOG_MESSAGE_NO_ID,
-            "This could be a file encoding issue. "
-            "Supported encodings must be backward compatible with ASCII.");
-        scanner_err(&scanner, XKB_LOG_MESSAGE_NO_ID,
-            "E.g. ISO/CEI 8859 and UTF-8 are supported "
-            "but UTF-16, UTF-32 and CP1026 are not.");
+        scanner_err(&scanner, XKB_ERROR_INVALID_FILE_ENCODING,
+                    "This could be a file encoding issue. "
+                    "Supported encodings must be backward compatible with ASCII.");
+        scanner_err(&scanner, XKB_ERROR_INVALID_FILE_ENCODING,
+                    "E.g. ISO/CEI 8859 and UTF-8 are supported "
+                    "but UTF-16, UTF-32 and CP1026 are not.");
         unmap_file(string, size);
         return false;
     }
@@ -1529,9 +1713,9 @@ xkb_components_from_rules(struct xkb_context *ctx,
 
     darray_steal(matcher->kccgst[KCCGST_KEYCODES], &out->keycodes, NULL);
     darray_steal(matcher->kccgst[KCCGST_TYPES], &out->types, NULL);
-    darray_steal(matcher->kccgst[KCCGST_COMPAT], &out->compat, NULL);
+    darray_steal(matcher->kccgst[KCCGST_COMPAT], &out->compatibility, NULL);
     darray_steal(matcher->kccgst[KCCGST_SYMBOLS], &out->symbols, NULL);
-    darray_free(matcher->kccgst[KCCGST_GEOMETRY]);
+    darray_steal(matcher->kccgst[KCCGST_GEOMETRY], &out->geometry, NULL);
 
     mval = &matcher->rmlvo.model;
     if (!mval->matched && mval->sval.len > 0)
@@ -1560,14 +1744,16 @@ xkb_components_from_rules(struct xkb_context *ctx,
         const char *symbols = out->symbols;
         /* Take the highest modifier */
         while ((symbols = strchr(symbols, ':')) != NULL && symbols[1] != '\0') {
-            char *endptr;
-            unsigned long group = strtoul(++symbols, &endptr, 10);
+            xkb_layout_index_t group = 0;
+            const int count = parse_dec_to_uint32_t(++symbols, SIZE_MAX, &group);
             /* Update only when valid group index, but continue parsing
              * even on invalid ones, as we do not handle them here. */
-            if (group > 0 && group <= XKB_MAX_GROUPS)
-                *explicit_layouts = MAX(*explicit_layouts,
-                                        (xkb_layout_index_t)group);
-            symbols = endptr;
+            if (count > 0 && (symbols[count] == '\0' ||
+                is_merge_mode_prefix(symbols[count])) &&
+                group > 0 && group <= XKB_MAX_GROUPS) {
+                *explicit_layouts = MAX(*explicit_layouts, group);
+                symbols += count;
+            }
         }
     }
 
